@@ -7,8 +7,8 @@ import urllib.parse
 import uuid
 from dataclasses import asdict
 
+import httpx
 import jwt
-import requests
 
 from . import ucjson
 from .constants import AUTHN_TYPE_PASSWORD
@@ -19,6 +19,8 @@ from .models import (
     AccessPolicyTemplate,
     APIErrorResponse,
     Column,
+    ColumnRetentionDurationResponse,
+    ColumnRetentionDurationsResponse,
     Edge,
     EdgeType,
     InspectTokenResponse,
@@ -29,22 +31,44 @@ from .models import (
     Purpose,
     ResourceID,
     Transformer,
+    UpdateColumnRetentionDurationRequest,
+    UpdateColumnRetentionDurationsRequest,
     UserResponse,
 )
 
 
-class Error(BaseException):
-    def __init__(self, error="unspecified error", code=500, request_id=None):
-        self.error = error
-        self.code = code
+class Error(Exception):
+    def __init__(
+        self,
+        error: str | dict = "unspecified error",
+        http_status_code: int = -1,
+        request_id: str | None = None,
+    ) -> None:
+        super().__init__(error)
+        self._err = error
+        self.error_json = error if isinstance(error, dict) else None
+        self.code = http_status_code
         self.request_id = request_id
 
     def __repr__(self):
-        return f"Error({self.error}, {self.code}, {self.request_id})"
+        return f"Error({self._err}, {self.code}, {self.request_id})"
 
-    @staticmethod
-    def from_json(j):
-        return Error(j["error"], j["request_id"])
+    @classmethod
+    def from_response(cls, resp) -> Error:
+        request_id = resp.headers.get("X-Request-Id")
+        if _is_json(resp):
+            resp_json = ucjson.loads(resp.text)
+            return cls(
+                error=resp_json["error"],
+                request_id=resp_json.get("request_id", request_id),
+                http_status_code=resp.status_code,
+            )
+        else:
+            return cls(
+                error=f"HTTP {resp.status_code} - {resp.text}",
+                request_id=request_id,
+                http_status_code=resp.status_code,
+            )
 
 
 def read_env(name: str, desc: str) -> str:
@@ -56,37 +80,53 @@ def read_env(name: str, desc: str) -> str:
     return value
 
 
-class Client:
-    url: str
-    client_id: str
-    _client_secret: str
-    _request_kwargs: dict
-    _access_token: str
+def create_http_client(url: str, **kwargs) -> httpx.Client:
+    return httpx.Client(base_url=url, **kwargs)
 
+
+def _is_json(resp) -> bool:
+    return resp.headers.get("Content-Type") == "application/json"
+
+
+class Client:
     @classmethod
-    def from_env(cls, **kwargs):
+    def from_env(cls, client_factory=create_http_client, **kwargs):
         return cls(
             url=read_env("TENANT_URL", "Tenant URL"),
-            id=read_env("CLIENT_ID", "Client ID"),
-            secret=read_env("CLIENT_SECRET", "Client Secret"),
+            client_id=read_env("CLIENT_ID", "Client ID"),
+            client_secret=read_env("CLIENT_SECRET", "Client Secret"),
+            client_factory=client_factory,
             **kwargs,
         )
 
-    def __init__(self, url: str, client_id: str, client_secret: str, **kwargs):
-        self.url = url
-        self.client_id = urllib.parse.quote(client_id)
-        self._client_secret = urllib.parse.quote(client_secret)
+    def __init__(
+        self,
+        url: str,
+        client_id: str,
+        client_secret: str,
+        client_factory=create_http_client,
+        **kwargs,
+    ):
+        self._authorization = base64.b64encode(
+            bytes(
+                f"{ urllib.parse.quote(client_id)}:{ urllib.parse.quote(client_secret)}",
+                "ISO-8859-1",
+            )
+        ).decode("ascii")
+        self._client = client_factory(url, **kwargs)
+        kwargs.pop(
+            "verify", None
+        )  # Remove verify from kwargs since it is not compatible with httpx request methods (get, post, etc...)
         self._request_kwargs = kwargs
-
-        self._access_token = self._get_access_token()
+        self._access_token: str | None = None  # lazy loaded
 
     # User Operations
 
-    def CreateUser(self) -> uuid.UUID:
-        body = {}
+    # AuthN user methods (shouldn't be used by UserStore customers)
 
-        j = self._post("/authn/users", data=ucjson.dumps(body))
-        return j.get("id")
+    def CreateUser(self) -> uuid.UUID:
+        resp_json = self._post("/authn/users", data=ucjson.dumps({}))
+        return resp_json.get("id")
 
     def CreateUserWithPassword(self, username: str, password: str) -> uuid.UUID:
         body = {
@@ -95,13 +135,16 @@ class Client:
             "authn_type": AUTHN_TYPE_PASSWORD,
         }
 
-        j = self._post("/authn/users", data=ucjson.dumps(body))
-        return j.get("id")
+        resp_json = self._post("/authn/users", data=ucjson.dumps(body))
+        return resp_json.get("id")
 
     def ListUsers(
-        self, limit: int = 0, starting_after: uuid.UUID = None, email: str = None
+        self,
+        limit: int = 0,
+        starting_after: uuid.UUID | None = None,
+        email: str | None = None,
     ) -> list[UserResponse]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -109,64 +152,82 @@ class Client:
         if email is not None:
             params["email"] = email
         params["version"] = "3"
-        j = self._get("/authn/users", params=params)
+        resp_json = self._get("/authn/users", params=params)
 
-        users = [UserResponse.from_json(ur) for ur in j["data"]]
+        users = [UserResponse.from_json(ur) for ur in resp_json["data"]]
         return users
 
     def GetUser(self, id: uuid.UUID) -> UserResponse:
-        j = self._get(f"/authn/users/{id}")
-        return UserResponse.from_json(j)
+        resp_json = self._get(f"/authn/users/{id}")
+        return UserResponse.from_json(resp_json)
 
     def UpdateUser(self, id: uuid.UUID, profile: dict) -> UserResponse:
         body = {"profile": profile}
-
-        j = self._put(f"/authn/users/{id}", data=ucjson.dumps(body))
-        return UserResponse.from_json(j)
+        resp_json = self._put(f"/authn/users/{id}", data=ucjson.dumps(body))
+        return UserResponse.from_json(resp_json)
 
     def DeleteUser(self, id: uuid.UUID) -> bool:
         return self._delete(f"/authn/users/{id}")
 
+    # Userstore user methods (should be used along with Accessor and Mutator methods)
+
+    def CreateUserWithMutator(
+        self,
+        mutator_id: uuid.UUID,
+        context: dict,
+        row_data: dict,
+        organization_id: uuid.UUID | None = None,
+    ) -> uuid.UUID:
+        body = {
+            "mutator_id": mutator_id,
+            "context": context,
+            "row_data": row_data,
+            "organization_id": organization_id,
+        }
+
+        j = self._post("/userstore/api/users", data=ucjson.dumps(body))
+        return j
+
     # Column Operations
 
     def CreateColumn(self, column: Column, if_not_exists=False) -> Column:
         body = {"column": column.__dict__}
-
         try:
-            j = self._post("/userstore/config/columns", data=ucjson.dumps(body))
-            return Column.from_json(j)
-        except Error as e:
+            resp_json = self._post("/userstore/config/columns", data=ucjson.dumps(body))
+            return Column.from_json(resp_json)
+        except Error as err:
             if if_not_exists:
-                column.id = _id_from_identical_conflict(e)
+                column.id = _id_from_identical_conflict(err)
                 return column
-            raise e
+            raise err
 
-    def DeleteColumn(self, id: uuid.UUID) -> str:
+    def DeleteColumn(self, id: uuid.UUID) -> bool:
         return self._delete(f"/userstore/config/columns/{id}")
 
     def GetColumn(self, id: uuid.UUID) -> Column:
-        j = self._get(f"/userstore/config/columns/{id}")
-        return Column.from_json(j)
+        resp_json = self._get(f"/userstore/config/columns/{id}")
+        return Column.from_json(resp_json)
 
     def ListColumns(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[Column]:
-        params = {}
+        params: dict[str, int | str] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
             params["starting_after"] = f"id:{starting_after}"
         params["version"] = "3"
-        j = self._get("/userstore/config/columns", params=params)
-
-        columns = [Column.from_json(c) for c in j["data"]]
+        resp_json = self._get("/userstore/config/columns", params=params)
+        columns = [Column.from_json(col) for col in resp_json["data"]]
         return columns
 
     def UpdateColumn(self, column: Column) -> Column:
         body = {"column": column.__dict__}
 
-        j = self._put(f"/userstore/config/columns/{column.id}", data=ucjson.dumps(body))
-        return Column.from_json(j)
+        resp_json = self._put(
+            f"/userstore/config/columns/{column.id}", data=ucjson.dumps(body)
+        )
+        return Column.from_json(resp_json)
 
     # Purpose Operations
 
@@ -174,93 +235,256 @@ class Client:
         body = {"purpose": purpose.__dict__}
 
         try:
-            j = self._post("/userstore/config/purposes", data=ucjson.dumps(body))
-            return Purpose.from_json(j)
-        except Error as e:
+            resp_json = self._post(
+                "/userstore/config/purposes", data=ucjson.dumps(body)
+            )
+            return Purpose.from_json(resp_json)
+        except Error as err:
             if if_not_exists:
-                purpose.id = _id_from_identical_conflict(e)
+                purpose.id = _id_from_identical_conflict(err)
                 return purpose
-            raise e
+            raise err
 
-    def DeletePurpose(self, id: uuid.UUID) -> str:
+    def DeletePurpose(self, id: uuid.UUID) -> bool:
         return self._delete(f"/userstore/config/purposes/{id}")
 
     def GetPurpose(self, id: uuid.UUID) -> Purpose:
-        j = self._get(f"/userstore/config/purposes/{id}")
-        return Purpose.from_json(j)
+        json_resp = self._get(f"/userstore/config/purposes/{id}")
+        return Purpose.from_json(json_resp)
 
     def ListPurposes(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[Purpose]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
             params["starting_after"] = f"id:{starting_after}"
         params["version"] = "3"
-        j = self._get("/userstore/config/purposes", params=params)
-
-        purposes = [Purpose.from_json(p) for p in j["data"]]
+        resp_json = self._get("/userstore/config/purposes", params=params)
+        purposes = [Purpose.from_json(p) for p in resp_json["data"]]
         return purposes
 
     def UpdatePurpose(self, purpose: Purpose) -> Purpose:
         body = {"purpose": purpose.__dict__}
 
-        j = self._put(
+        resp_json = self._put(
             f"/userstore/config/purposes/{purpose.id}", data=ucjson.dumps(body)
         )
-        return Purpose.from_json(j)
+        return Purpose.from_json(resp_json)
+
+    # Retention Duration Operations
+
+    # Tenant Retention Duration
+
+    # A configured tenant retention duration will apply for
+    # all column purposes that do not have a configured purpose
+    # retention duration default or a configured column purpose
+    # retention duration. If a tenant retention duration is
+    # not configured, soft-deleted values will not be retained
+    # by default.
+
+    # create a tenant retention duration default
+    def CreateSoftDeletedRetentionDurationOnTenant(
+        self, req: UpdateColumnRetentionDurationRequest
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._post(
+            "/userstore/config/softdeletedretentiondurations", data=req.to_json()
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # delete a tenant retention duration default
+    def DeleteSoftDeletedRetentionDurationOnTenant(self, durationID: uuid.UUID) -> bool:
+        return self._delete(
+            f"/userstore/config/softdeletedretentiondurations/{durationID}"
+        )
+
+    # get a specific tenant retention duration default
+    def GetSoftDeletedRetentionDurationOnTenant(
+        self, durationID: uuid.UUID
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._get(
+            f"/userstore/config/softdeletedretentiondurations/{durationID}"
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # get tenant retention duration, or default value if not specified
+    def GetDefaultSoftDeletedRetentionDurationOnTenant(
+        self,
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._get("/userstore/config/softdeletedretentiondurations")
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # update a specific tenant retention duration default
+    def UpdateSoftDeletedRetentionDurationOnTenant(
+        self, durationID: uuid.UUID, req: UpdateColumnRetentionDurationRequest
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._put(
+            f"/userstore/config/softdeletedretentiondurations/{durationID}",
+            req.to_json(),
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # Purpose Retention Durations
+
+    # A configured purpose retention duration will apply for all
+    # column purposes that include the specified purpose, unless
+    # a retention duration has been configured for a specific
+    # column purpose.
+
+    # create a purpose retention duration default
+    def CreateSoftDeletedRetentionDurationOnPurpose(
+        self, purposeID: uuid.UUID, req: UpdateColumnRetentionDurationRequest
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._post(
+            f"/userstore/config/purposes/{purposeID}/softdeletedretentiondurations",
+            data=req.to_json(),
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # delete a purpose retention duration default
+    def DeleteSoftDeletedRetentionDurationOnPurpose(
+        self, purposeID: uuid.UUID, durationID: uuid.UUID
+    ) -> bool:
+        return self._delete(
+            f"/userstore/config/purposes/{purposeID}/softdeletedretentiondurations/{durationID}"
+        )
+
+    # get a specific purpose retention duration default
+    def GetSoftDeletedRetentionDurationOnPurpose(
+        self, purposeID: uuid.UUID, durationID: uuid.UUID
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._get(
+            f"/userstore/config/purposes/{purposeID}/softdeletedretentiondurations/{durationID}"
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # get purpose retention duration, or default value if not specified
+    def GetDefaultSoftDeletedRetentionDurationOnPurpose(
+        self, purposeID: uuid.UUID
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._get(
+            f"/userstore/config/purposes/{purposeID}/softdeletedretentiondurations"
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # update a specific purpose retention duration default
+    def UpdateSoftDeletedRetentionDurationOnPurpose(
+        self,
+        purposeID: uuid.UUID,
+        durationID: uuid.UUID,
+        req: UpdateColumnRetentionDurationRequest,
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._put(
+            f"/userstore/config/purposes/{purposeID}/softdeletedretentiondurations/{durationID}",
+            req.to_json(),
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # Column Retention Durations
+
+    # A configured column purpose retention duration will override
+    # any configured purpose level, tenant level, or system-level
+    # default retention durations.
+
+    # get a specific column purpose retention duration
+    def GetSoftDeletedRetentionDurationOnColumn(
+        self, columnID: uuid.UUID, durationID: uuid.UUID
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._get(
+            f"/userstore/config/columns/{columnID}/softdeletedretentiondurations/{durationID}"
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # get the retention duration for each purpose for a given column
+    def GetSoftDeletedRetentionDurationsOnColumn(
+        self, columnID: uuid.UUID
+    ) -> ColumnRetentionDurationsResponse:
+        resp = self._get(
+            f"/userstore/config/columns/{columnID}/softdeletedretentiondurations"
+        )
+        return ColumnRetentionDurationsResponse.from_json(resp)
+
+    # delete a specific column purpose retention duration
+    def DeleteSoftDeletedRetentionDurationOnColumn(
+        self, columnID: uuid.UUID, durationID: uuid.UUID
+    ) -> bool:
+        return self._delete(
+            f"/userstore/config/columns/{columnID}/softdeletedretentiondurations/{durationID}"
+        )
+
+    # update a specific column purpose retention duration
+    def UpdateSoftDeletedRetentionDurationOnColumn(
+        self,
+        columnID: uuid.UUID,
+        durationID: uuid.UUID,
+        req: UpdateColumnRetentionDurationRequest,
+    ) -> ColumnRetentionDurationResponse:
+        resp = self._put(
+            f"/userstore/config/columns/{columnID}/softdeletedretentiondurations/{durationID}",
+            req.to_json(),
+        )
+        return ColumnRetentionDurationResponse.from_json(resp)
+
+    # update the specified purpose retention durations for the column
+    # - durations can be added, deleted, or updated for each purpose
+    def UpdateSoftDeletedRetentionDurationsOnColumn(
+        self, columnID: uuid.UUID, req: UpdateColumnRetentionDurationsRequest
+    ) -> ColumnRetentionDurationsResponse:
+        resp = self._post(
+            f"/userstore/config/columns/{columnID}/softdeletedretentiondurations",
+            data=req.to_json(),
+        )
+        return ColumnRetentionDurationsResponse.from_json(resp)
 
     # Access Policy Templates
 
-    # Access Policies
-
     def CreateAccessPolicyTemplate(
         self, access_policy_template: AccessPolicyTemplate, if_not_exists=False
     ) -> AccessPolicyTemplate | Error:
         body = {"access_policy_template": access_policy_template.__dict__}
 
         try:
-            j = self._post(
+            resp_json = self._post(
                 "/tokenizer/policies/accesstemplate", data=ucjson.dumps(body)
             )
-            return AccessPolicyTemplate.from_json(j)
-        except Error as e:
+            return AccessPolicyTemplate.from_json(resp_json)
+        except Error as err:
             if if_not_exists:
-                access_policy_template.id = _id_from_identical_conflict(e)
+                access_policy_template.id = _id_from_identical_conflict(err)
                 return access_policy_template
-            raise e
+            raise err
 
     def ListAccessPolicyTemplates(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ):
-        params = {}
+        params: dict[str, int | str] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
             params["starting_after"] = f"id:{starting_after}"
         params["version"] = "3"
-        j = self._get("/tokenizer/policies/accesstemplate", params=params)
+        resp_json = self._get("/tokenizer/policies/accesstemplate", params=params)
 
-        templates = [AccessPolicyTemplate.from_json(p) for p in j["data"]]
+        templates = [AccessPolicyTemplate.from_json(apt) for apt in resp_json["data"]]
         return templates
 
     def GetAccessPolicyTemplate(self, rid: ResourceID):
         if rid.id is not None:
-            j = self._get(f"/tokenizer/policies/accesstemplate/{rid.id}")
+            resp_json = self._get(f"/tokenizer/policies/accesstemplate/{rid.id}")
         elif rid.name is not None:
-            j = self._get(f"/tokenizer/policies/accesstemplate?name={rid.name}")
+            resp_json = self._get(f"/tokenizer/policies/accesstemplate?name={rid.name}")
 
-        return AccessPolicyTemplate.from_json(j)
+        return AccessPolicyTemplate.from_json(resp_json)
 
     def UpdateAccessPolicyTemplate(self, access_policy_template: AccessPolicyTemplate):
         body = {"access_policy_template": access_policy_template.__dict__}
 
-        j = self._put(
+        resp_json = self._put(
             f"/tokenizer/policies/accesstemplate/{access_policy_template.id}",
             data=ucjson.dumps(body),
         )
-        return AccessPolicyTemplate.from_json(j)
+        return AccessPolicyTemplate.from_json(resp_json)
 
     def DeleteAccessPolicyTemplate(self, id: uuid.UUID, version: int):
         return self._delete(
@@ -276,42 +500,46 @@ class Client:
         body = {"access_policy": access_policy.__dict__}
 
         try:
-            j = self._post("/tokenizer/policies/access", data=ucjson.dumps(body))
-            return AccessPolicy.from_json(j)
-        except Error as e:
+            resp_json = self._post(
+                "/tokenizer/policies/access", data=ucjson.dumps(body)
+            )
+            return AccessPolicy.from_json(resp_json)
+        except Error as err:
             if if_not_exists:
-                access_policy.id = _id_from_identical_conflict(e)
+                access_policy.id = _id_from_identical_conflict(err)
                 return access_policy
-            raise e
+            raise err
 
-    def ListAccessPolicies(self, limit: int = 0, starting_after: uuid.UUID = None):
-        params = {}
+    def ListAccessPolicies(
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
+    ):
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
             params["starting_after"] = f"id:{starting_after}"
         params["version"] = "3"
-        j = self._get("/tokenizer/policies/access", params=params)
+        resp_json = self._get("/tokenizer/policies/access", params=params)
 
-        policies = [AccessPolicy.from_json(p) for p in j["data"]]
+        policies = [AccessPolicy.from_json(ap) for ap in resp_json["data"]]
         return policies
 
     def GetAccessPolicy(self, rid: ResourceID):
         if rid.id is not None:
-            j = self._get(f"/tokenizer/policies/access/{rid.id}")
+            resp_json = self._get(f"/tokenizer/policies/access/{rid.id}")
         elif rid.name is not None:
-            j = self._get(f"/tokenizer/policies/access?name={rid.name}")
+            resp_json = self._get(f"/tokenizer/policies/access?name={rid.name}")
 
-        return AccessPolicy.from_json(j)
+        return AccessPolicy.from_json(resp_json)
 
     def UpdateAccessPolicy(self, access_policy: AccessPolicy):
         body = {"access_policy": access_policy.__dict__}
 
-        j = self._put(
+        resp_json = self._put(
             f"/tokenizer/policies/access/{access_policy.id}",
             data=ucjson.dumps(body),
         )
-        return AccessPolicy.from_json(j)
+        return AccessPolicy.from_json(resp_json)
 
     def DeleteAccessPolicy(self, id: uuid.UUID, version: int):
         return self._delete(
@@ -319,32 +547,31 @@ class Client:
             params={"policy_version": str(version)},
         )
 
-    ### Transformers
+    # Transformers
 
     def CreateTransformer(self, transformer: Transformer, if_not_exists=False):
         body = {"transformer": transformer.__dict__}
 
         try:
-            j = self._post(
+            resp_json = self._post(
                 "/tokenizer/policies/transformation", data=ucjson.dumps(body)
             )
-            return Transformer.from_json(j)
-        except Error as e:
+            return Transformer.from_json(resp_json)
+        except Error as err:
             if if_not_exists:
-                transformer.id = _id_from_identical_conflict(e)
+                transformer.id = _id_from_identical_conflict(err)
                 return transformer
-            raise e
+            raise err
 
-    def ListTransformers(self, limit: int = 0, starting_after: uuid.UUID = None):
-        params = {}
+    def ListTransformers(self, limit: int = 0, starting_after: uuid.UUID | None = None):
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
             params["starting_after"] = f"id:{starting_after}"
         params["version"] = "3"
-        j = self._get("/tokenizer/policies/transformation", params=params)
-
-        transformers = [Transformer.from_json(p) for p in j["data"]]
+        resp_json = self._get("/tokenizer/policies/transformation", params=params)
+        transformers = [Transformer.from_json(tf) for tf in resp_json["data"]]
         return transformers
 
     # Note: Transformers are immutable, so no Update method is provided.
@@ -358,15 +585,17 @@ class Client:
         body = {"accessor": accessor.__dict__}
 
         try:
-            j = self._post("/userstore/config/accessors", data=ucjson.dumps(body))
-            return Accessor.from_json(j)
-        except Error as e:
+            resp_json = self._post(
+                "/userstore/config/accessors", data=ucjson.dumps(body)
+            )
+            return Accessor.from_json(resp_json)
+        except Error as err:
             if if_not_exists:
-                accessor.id = _id_from_identical_conflict(e)
+                accessor.id = _id_from_identical_conflict(err)
                 return accessor
-            raise e
+            raise err
 
-    def DeleteAccessor(self, id: uuid.UUID) -> str:
+    def DeleteAccessor(self, id: uuid.UUID) -> bool:
         return self._delete(f"/userstore/config/accessors/{id}")
 
     def GetAccessor(self, id: uuid.UUID) -> Accessor:
@@ -374,27 +603,27 @@ class Client:
         return Accessor.from_json(j)
 
     def ListAccessors(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[Accessor]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
             params["starting_after"] = f"id:{starting_after}"
         params["version"] = "3"
-        j = self._get("/userstore/config/accessors", params=params)
+        resp_json = self._get("/userstore/config/accessors", params=params)
 
-        accessors = [Accessor.from_json(a) for a in j["data"]]
+        accessors = [Accessor.from_json(acs) for acs in resp_json["data"]]
         return accessors
 
     def UpdateAccessor(self, accessor: Accessor) -> Accessor:
         body = {"accessor": accessor.__dict__}
 
-        j = self._put(
+        resp_json = self._put(
             f"/userstore/config/accessors/{accessor.id}",
             data=ucjson.dumps(body),
         )
-        return Accessor.from_json(j)
+        return Accessor.from_json(resp_json)
 
     def ExecuteAccessor(
         self, accessor_id: uuid.UUID, context: dict, selector_values: list
@@ -408,6 +637,7 @@ class Client:
         return self._post("/userstore/api/accessors", data=ucjson.dumps(body))
 
     # Mutator Operations
+
     def CreateMutator(self, mutator: Mutator, if_not_exists=False) -> Mutator:
         body = {"mutator": mutator.__dict__}
 
@@ -420,7 +650,7 @@ class Client:
                 return mutator
             raise e
 
-    def DeleteMutator(self, id: uuid.UUID) -> str:
+    def DeleteMutator(self, id: uuid.UUID) -> bool:
         return self._delete(f"/userstore/config/mutators/{id}")
 
     def GetMutator(self, id: uuid.UUID) -> Mutator:
@@ -428,9 +658,9 @@ class Client:
         return Mutator.from_json(j)
 
     def ListMutators(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[Mutator]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -467,7 +697,7 @@ class Client:
         j = self._post("/userstore/api/mutators", data=ucjson.dumps(body))
         return j
 
-    ### Token Operations
+    # Token Operations
 
     def CreateToken(
         self,
@@ -481,8 +711,8 @@ class Client:
             "access_policy_rid": access_policy_rid.__dict__,
         }
 
-        j = self._post("/tokenizer/tokens", data=ucjson.dumps(body))
-        return j["data"]
+        json_resp = self._post("/tokenizer/tokens", data=ucjson.dumps(body))
+        return json_resp["data"]
 
     def LookupOrCreateTokens(
         self,
@@ -536,9 +766,9 @@ class Client:
     # AuthZ Operations
 
     def ListObjects(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[Object]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -568,8 +798,10 @@ class Client:
     def DeleteObject(self, id: uuid.UUID):
         return self._delete(f"/authz/objects/{id}")
 
-    def ListEdges(self, limit: int = 0, starting_after: uuid.UUID = None) -> list[Edge]:
-        params = {}
+    def ListEdges(
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
+    ) -> list[Edge]:
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -600,9 +832,9 @@ class Client:
         return self._delete(f"/authz/edges/{id}")
 
     def ListObjectTypes(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[ObjectType]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -635,9 +867,9 @@ class Client:
         return self._delete(f"/authz/objecttypes/{id}")
 
     def ListEdgeTypes(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[EdgeType]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -664,13 +896,13 @@ class Client:
         j = self._get(f"/authz/edgetypes/{id}")
         return EdgeType.from_json(j)
 
-    def DeleteEdgeType(self, id: uuid.UUID):
+    def DeleteEdgeType(self, id: uuid.UUID) -> bool:
         return self._delete(f"/authz/edgetypes/{id}")
 
     def ListOrganizations(
-        self, limit: int = 0, starting_after: uuid.UUID = None
+        self, limit: int = 0, starting_after: uuid.UUID | None = None
     ) -> list[Organization]:
-        params = {}
+        params: dict[str, str | int] = {}
         if limit > 0:
             params["limit"] = limit
         if starting_after is not None:
@@ -683,14 +915,14 @@ class Client:
 
     def CreateOrganization(self, organization: Organization) -> Organization:
         body = organization.__dict__
-        j = self._post("/authz/organizations", data=ucjson.dumps(body))
-        return Organization.from_json(j)
+        json_data = self._post("/authz/organizations", data=ucjson.dumps(body))
+        return Organization.from_json(json_data)
 
     def GetOrganization(self, id: uuid.UUID) -> Organization:
-        j = self._get(f"/authz/organizations/{id}")
-        return Organization.from_json(j)
+        json_data = self._get(f"/authz/organizations/{id}")
+        return Organization.from_json(json_data)
 
-    def DeleteOrganization(self, id: uuid.UUID):
+    def DeleteOrganization(self, id: uuid.UUID) -> bool:
         return self._delete(f"/authz/organizations/{id}")
 
     def CheckAttribute(
@@ -704,33 +936,35 @@ class Client:
         )
         return j.get("has_attribute")
 
-    # Access token helpers
+    def DownloadUserstoreSDK(self) -> str:
+        return self._download("/userstore/download/codegensdk.py")
+
+    # Access Token Helpers
 
     def _get_access_token(self) -> str:
         # Encode the client ID and client secret
-        authorization = base64.b64encode(
-            bytes(f"{self.client_id}:{self._client_secret}", "ISO-8859-1")
-        ).decode("ascii")
-
         headers = {
-            "Authorization": f"Basic {authorization}",
+            "Authorization": f"Basic {self._authorization}",
             "Content-Type": "application/x-www-form-urlencoded",
         }
         body = {"grant_type": "client_credentials"}
 
         # Note that we use requests directly here (instead of _post) because we don't
         # want to refresh the access token as we are trying to get it. :)
-        r = requests.post(
-            self.url + "/oidc/token",
+        resp = self._client.post(
+            "/oidc/token",
             headers=headers,
             data=body,
             **self._request_kwargs,
         )
-        j = ucjson.loads(r.text)
-        return j.get("access_token")
+        if resp.status_code >= 400:
+            raise Error.from_response(resp)
+        json_data = ucjson.loads(resp.text)
+        return json_data.get("access_token")
 
-    def _refresh_access_token_if_needed(self):
+    def _refresh_access_token_if_needed(self) -> None:
         if self._access_token is None:
+            self._access_token = self._get_access_token()
             return
 
         # TODO: this takes advantage of an implementation detail that we use JWTs for
@@ -748,68 +982,59 @@ class Client:
     def _get_headers(self) -> dict:
         return {"Authorization": f"Bearer {self._access_token}"}
 
-    # Request helpers
+    # Request Helpers
 
     def _get(self, url, **kwargs) -> dict:
         self._refresh_access_token_if_needed()
         args = self._request_kwargs.copy()
         args.update(kwargs)
-        r = requests.get(self.url + url, headers=self._get_headers(), **args)
-        j = ucjson.loads(r.text)
+        resp = self._client.get(url, headers=self._get_headers(), **args)
+        if resp.status_code >= 400:
+            raise Error.from_response(resp)
+        return ucjson.loads(resp.text)
 
-        if r.status_code >= 400:
-            e = Error.from_json(j)
-            e.code = r.status_code
-            raise e
-
-        return j
-
-    def _post(self, url, **kwargs) -> dict:
+    def _post(self, url, **kwargs) -> dict | list:
         self._refresh_access_token_if_needed()
         args = self._request_kwargs.copy()
         args.update(kwargs)
-        r = requests.post(self.url + url, headers=self._get_headers(), **args)
-        j = ucjson.loads(r.text)
+        resp = self._client.post(url, headers=self._get_headers(), **args)
+        if resp.status_code >= 400:
+            raise Error.from_response(resp)
+        return ucjson.loads(resp.text)
 
-        if r.status_code >= 400:
-            e = Error.from_json(j)
-            e.code = r.status_code
-            raise e
-
-        return j
-
-    def _put(self, url, **kwargs) -> dict:
+    def _put(self, url, **kwargs) -> dict | list:
         self._refresh_access_token_if_needed()
         args = self._request_kwargs.copy()
         args.update(kwargs)
-        r = requests.put(self.url + url, headers=self._get_headers(), **args)
-        j = ucjson.loads(r.text)
-
-        if r.status_code >= 400:
-            e = Error.from_json(j)
-            e.code = r.status_code
-            raise e
-
-        return j
+        resp = self._client.put(url, headers=self._get_headers(), **args)
+        if resp.status_code >= 400:
+            raise Error.from_response(resp)
+        return ucjson.loads(resp.text)
 
     def _delete(self, url, **kwargs) -> bool:
         self._refresh_access_token_if_needed()
         args = self._request_kwargs.copy()
         args.update(kwargs)
-        r = requests.delete(self.url + url, headers=self._get_headers(), **args)
+        resp = self._client.delete(url, headers=self._get_headers(), **args)
 
-        if r.status_code >= 400:
-            j = ucjson.loads(r.text)
-            e = Error.from_json(j)
-            e.code = r.status_code
-            raise e
+        if resp.status_code == 404:
+            return False
 
-        return r.status_code == 204
+        if resp.status_code >= 400:
+            raise Error.from_response(resp)
+        return resp.status_code == 204
+
+    def _download(self, url, **kwargs) -> str:
+        self._refresh_access_token_if_needed()
+        args = self._request_kwargs.copy()
+        args.update(kwargs)
+        resp = self._client.get(url, headers=self._get_headers(), **args)
+        return resp.text
 
 
-def _id_from_identical_conflict(e):
-    if e.code == 409:
-        er = APIErrorResponse.from_json(e.error)
-        if er.identical:
-            return er.id
-    raise e
+def _id_from_identical_conflict(err: Error) -> uuid.UUID:
+    if err.code == 409:
+        api_error = APIErrorResponse.from_json(err.error_json)
+        if api_error.identical:
+            return api_error.id
+    raise err
